# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Initially imported from https://github.com/DataDog/datadogpy/
# at revision 62b3a3e89988dc18d78c282fe3ff5d1813917436
#
# Copyright (c) 2015, Datadog <info@datadoghq.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Datadog nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# Vastly adapted for integration in swh.core:
#
# - Removed python < 3.5 compat code
# - trimmed the imports down to be a single module
# - adjust some options:
#   - drop unix socket connection option
#   - add environment variable support for setting the statsd host and
#     port (pulled the idea from the main python statsd module)
#   - only send timer metrics in milliseconds (that's what
#     prometheus-statsd-exporter expects)
#   - drop DataDog-specific metric types (that are unsupported in
#     prometheus-statsd-exporter)
# - made the tags a dict instead of a list (prometheus-statsd-exporter only
#   supports tags with a value, mirroring prometheus)
# - switch from time.time to time.monotonic
# - improve unit test coverage
# - documentation cleanup


from asyncio import iscoroutinefunction
from functools import wraps
from random import random
from time import monotonic
import itertools
import logging
import os
import socket
import warnings


log = logging.getLogger('swh.core.statsd')


class TimedContextManagerDecorator(object):
    """
    A context manager and a decorator which will report the elapsed time in
    the context OR in a function call.

    Attributes:
      elapsed (float): the elapsed time at the point of completion
    """
    def __init__(self, statsd, metric=None, error_metric=None,
                 tags=None, sample_rate=1):
        self.statsd = statsd
        self.metric = metric
        self.error_metric = error_metric
        self.tags = tags
        self.sample_rate = sample_rate
        self.elapsed = None  # this is for testing purpose

    def __call__(self, func):
        """
        Decorator which returns the elapsed time of the function call.

        Default to the function name if metric was not provided.
        """
        if not self.metric:
            self.metric = '%s.%s' % (func.__module__, func.__name__)

        # Coroutines
        if iscoroutinefunction(func):
            @wraps(func)
            async def wrapped_co(*args, **kwargs):
                start = monotonic()
                try:
                    result = await func(*args, **kwargs)
                except:  # noqa
                    self._send_error()
                    raise
                self._send(start)
                return result
            return wrapped_co

        # Others
        @wraps(func)
        def wrapped(*args, **kwargs):
            start = monotonic()
            try:
                result = func(*args, **kwargs)
            except:  # noqa
                self._send_error()
                raise
            self._send(start)
            return result
        return wrapped

    def __enter__(self):
        if not self.metric:
            raise TypeError("Cannot used timed without a metric!")
        self._start = monotonic()
        return self

    def __exit__(self, type, value, traceback):
        # Report the elapsed time of the context manager if no error.
        if type is None:
            self._send(self._start)
        else:
            self._send_error()

    def _send(self, start):
        elapsed = (monotonic() - start) * 1000
        self.statsd.timing(self.metric, elapsed,
                           tags=self.tags, sample_rate=self.sample_rate)
        self.elapsed = elapsed

    def _send_error(self):
        if self.error_metric is None:
            self.error_metric = self.metric + '_error_count'
        self.statsd.increment(self.error_metric, tags=self.tags)

    def start(self):
        """Start the timer"""
        self.__enter__()

    def stop(self):
        """Stop the timer, send the metric value"""
        self.__exit__(None, None, None)


class Statsd(object):
    """Initialize a client to send metrics to a StatsD server.

    Arguments:
      host (str): the host of the StatsD server. Defaults to localhost.
      port (int): the port of the StatsD server. Defaults to 8125.

      max_buffer_size (int): Maximum number of metrics to buffer before
        sending to the server if sending metrics in batch

      namespace (str): Namespace to prefix all metric names

      constant_tags (Dict[str, str]): Tags to attach to all metrics

    Note:
      This class also supports the following environment variables:

      STATSD_HOST
        Override the default host of the statsd server
      STATSD_PORT
        Override the default port of the statsd server
      STATSD_TAGS
        Tags to attach to every metric reported. Example value:

        "label:value,other_label:other_value"
    """

    def __init__(self, host=None, port=None, max_buffer_size=50,
                 namespace=None, constant_tags=None):
        # Connection
        if host is None:
            host = os.environ.get('STATSD_HOST') or 'localhost'
        self.host = host

        if port is None:
            port = os.environ.get('STATSD_PORT') or 8125
        self.port = int(port)

        # Socket
        self.socket = None
        self.max_buffer_size = max_buffer_size
        self._send = self._send_to_server
        self.encoding = 'utf-8'

        # Tags
        self.constant_tags = {}
        tags_envvar = os.environ.get('STATSD_TAGS', '')
        for tag in tags_envvar.split(','):
            if not tag:
                continue
            if ':' not in tag:
                warnings.warn(
                    'STATSD_TAGS needs to be in key:value format, '
                    '%s invalid' % tag,
                    UserWarning,
                )
                continue
            k, v = tag.split(':', 1)
            self.constant_tags[k] = v

        if constant_tags:
            self.constant_tags.update({
                str(k): str(v)
                for k, v in constant_tags.items()
            })

        # Namespace
        if namespace is not None:
            namespace = str(namespace)
        self.namespace = namespace

    def __enter__(self):
        self.open_buffer(self.max_buffer_size)
        return self

    def __exit__(self, type, value, traceback):
        self.close_buffer()

    def gauge(self, metric, value, tags=None, sample_rate=1):
        """
        Record the value of a gauge, optionally setting a list of tags and a
        sample rate.

        >>> statsd.gauge('users.online', 123)
        >>> statsd.gauge('active.connections', 1001, tags={"protocol": "http"})
        """
        return self._report(metric, 'g', value, tags, sample_rate)

    def increment(self, metric, value=1, tags=None, sample_rate=1):
        """
        Increment a counter, optionally setting a value, tags and a sample
        rate.

        >>> statsd.increment('page.views')
        >>> statsd.increment('files.transferred', 124)
        """
        self._report(metric, 'c', value, tags, sample_rate)

    def decrement(self, metric, value=1, tags=None, sample_rate=1):
        """
        Decrement a counter, optionally setting a value, tags and a sample
        rate.

        >>> statsd.decrement('files.remaining')
        >>> statsd.decrement('active.connections', 2)
        """
        metric_value = -value if value else value
        self._report(metric, 'c', metric_value, tags, sample_rate)

    def histogram(self, metric, value, tags=None, sample_rate=1):
        """
        Sample a histogram value, optionally setting tags and a sample rate.

        >>> statsd.histogram('uploaded.file.size', 1445)
        >>> statsd.histogram('file.count', 26, tags={"filetype": "python"})
        """
        self._report(metric, 'h', value, tags, sample_rate)

    def timing(self, metric, value, tags=None, sample_rate=1):
        """
        Record a timing, optionally setting tags and a sample rate.

        >>> statsd.timing("query.response.time", 1234)
        """
        self._report(metric, 'ms', value, tags, sample_rate)

    def timed(self, metric=None, error_metric=None, tags=None, sample_rate=1):
        """
        A decorator or context manager that will measure the distribution of a
        function's/context's run time. Optionally specify a list of tags or a
        sample rate. If the metric is not defined as a decorator, the module
        name and function name will be used. The metric is required as a
        context manager.
        ::

            @statsd.timed('user.query.time', sample_rate=0.5)
            def get_user(user_id):
                # Do what you need to ...
                pass

            # Is equivalent to ...
            with statsd.timed('user.query.time', sample_rate=0.5):
                # Do what you need to ...
                pass

            # Is equivalent to ...
            start = time.monotonic()
            try:
                get_user(user_id)
            finally:
                statsd.timing('user.query.time', time.monotonic() - start)
        """
        return TimedContextManagerDecorator(
            statsd=self, metric=metric,
            error_metric=error_metric,
            tags=tags, sample_rate=sample_rate)

    def set(self, metric, value, tags=None, sample_rate=1):
        """
        Sample a set value.

        >>> statsd.set('visitors.uniques', 999)
        """
        self._report(metric, 's', value, tags, sample_rate)

    def get_socket(self):
        """
        Return a connected socket.

        Note: connect the socket before assigning it to the class instance to
        avoid bad thread race conditions.
        """
        if not self.socket:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect((self.host, self.port))
            self.socket = sock

        return self.socket

    def open_buffer(self, max_buffer_size=50):
        """
        Open a buffer to send a batch of metrics in one packet.

        You can also use this as a context manager.

        >>> with Statsd() as batch:
        ...     batch.gauge('users.online', 123)
        ...     batch.gauge('active.connections', 1001)
        """
        self.max_buffer_size = max_buffer_size
        self.buffer = []
        self._send = self._send_to_buffer

    def close_buffer(self):
        """
        Flush the buffer and switch back to single metric packets.
        """
        self._send = self._send_to_server

        if self.buffer:
            # Only send packets if there are packets to send
            self._flush_buffer()

    def close_socket(self):
        """
        Closes connected socket if connected.
        """
        if self.socket:
            self.socket.close()
            self.socket = None

    def _report(self, metric, metric_type, value, tags, sample_rate):
        """
        Create a metric packet and send it.
        """
        if value is None:
            return

        if sample_rate != 1 and random() > sample_rate:
            return

        # Resolve the full tag list
        tags = self._add_constant_tags(tags)

        # Create/format the metric packet
        payload = "%s%s:%s|%s%s%s" % (
            (self.namespace + ".") if self.namespace else "",
            metric,
            value,
            metric_type,
            ("|@" + str(sample_rate)) if sample_rate != 1 else "",
            ("|#" + ",".join(
                "%s:%s" % (k, v)
                for (k, v) in sorted(tags.items())
            )) if tags else "",
        )
        # Send it
        self._send(payload)

    def _send_to_server(self, packet):
        try:
            # If set, use socket directly
            (self.socket or self.get_socket()).send(packet.encode('utf-8'))
        except socket.timeout:
            return
        except socket.error:
            log.debug(
                "Error submitting statsd packet."
                " Dropping the packet and closing the socket."
            )
            self.close_socket()

    def _send_to_buffer(self, packet):
        self.buffer.append(packet)
        if len(self.buffer) >= self.max_buffer_size:
            self._flush_buffer()

    def _flush_buffer(self):
        self._send_to_server("\n".join(self.buffer))
        self.buffer = []

    def _add_constant_tags(self, tags):
        return {
            str(k): str(v)
            for k, v in itertools.chain(
                    self.constant_tags.items(),
                    (tags if tags else {}).items(),
            )
        }


statsd = Statsd()



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import binascii
import datetime
import enum
import json
import os
import threading

from contextlib import contextmanager

import psycopg2
import psycopg2.extras


psycopg2.extras.register_uuid()


def escape(data):
    if data is None:
        return ''
    if isinstance(data, bytes):
        return '\\x%s' % binascii.hexlify(data).decode('ascii')
    elif isinstance(data, str):
        return '"%s"' % data.replace('"', '""')
    elif isinstance(data, datetime.datetime):
        # We escape twice to make sure the string generated by
        # isoformat gets escaped
        return escape(data.isoformat())
    elif isinstance(data, dict):
        return escape(json.dumps(data))
    elif isinstance(data, list):
        return escape("{%s}" % ','.join(escape(d) for d in data))
    elif isinstance(data, psycopg2.extras.Range):
        # We escape twice here too, so that we make sure
        # everything gets passed to copy properly
        return escape(
            '%s%s,%s%s' % (
                '[' if data.lower_inc else '(',
                '-infinity' if data.lower_inf else escape(data.lower),
                'infinity' if data.upper_inf else escape(data.upper),
                ']' if data.upper_inc else ')',
            )
        )
    elif isinstance(data, enum.IntEnum):
        return escape(int(data))
    else:
        # We don't escape here to make sure we pass literals properly
        return str(data)


def typecast_bytea(value, cur):
    if value is not None:
        data = psycopg2.BINARY(value, cur)
        return data.tobytes()


class BaseDb:
    """Base class for swh.*.*Db.

    cf. swh.storage.db.Db, swh.archiver.db.ArchiverDb

    """

    @classmethod
    def adapt_conn(cls, conn):
        """Makes psycopg2 use 'bytes' to decode bytea instead of
        'memoryview', for this connection."""
        cur = conn.cursor()
        cur.execute("SELECT null::bytea, null::bytea[]")
        bytea_oid = cur.description[0][1]
        bytea_array_oid = cur.description[1][1]

        t_bytes = psycopg2.extensions.new_type(
            (bytea_oid,), "bytea", typecast_bytea)
        psycopg2.extensions.register_type(t_bytes, conn)

        t_bytes_array = psycopg2.extensions.new_array_type(
            (bytea_array_oid,), "bytea[]", t_bytes)
        psycopg2.extensions.register_type(t_bytes_array, conn)

    @classmethod
    def connect(cls, *args, **kwargs):
        """factory method to create a DB proxy

        Accepts all arguments of psycopg2.connect; only some specific
        possibilities are reported below.

        Args:
            connstring: libpq2 connection string

        """
        conn = psycopg2.connect(*args, **kwargs)
        return cls(conn)

    @classmethod
    def from_pool(cls, pool):
        conn = pool.getconn()
        return cls(conn, pool=pool)

    def __init__(self, conn, pool=None):
        """create a DB proxy

        Args:
            conn: psycopg2 connection to the SWH DB
            pool: psycopg2 pool of connections

        """
        self.adapt_conn(conn)
        self.conn = conn
        self.pool = pool

    def put_conn(self):
        if self.pool:
            self.pool.putconn(self.conn)

    def cursor(self, cur_arg=None):
        """get a cursor: from cur_arg if given, or a fresh one otherwise

        meant to avoid boilerplate if/then/else in methods that proxy stored
        procedures

        """
        if cur_arg is not None:
            return cur_arg
        else:
            return self.conn.cursor()
    _cursor = cursor  # for bw compat

    @contextmanager
    def transaction(self):
        """context manager to execute within a DB transaction

        Yields:
            a psycopg2 cursor

        """
        with self.conn.cursor() as cur:
            try:
                yield cur
                self.conn.commit()
            except Exception:
                if not self.conn.closed:
                    self.conn.rollback()
                raise

    def copy_to(self, items, tblname, columns,
                cur=None, item_cb=None, default_values={}):
        """Copy items' entries to table tblname with columns information.

        Args:
            items (List[dict]): dictionaries of data to copy over tblname.
            tblname (str): destination table's name.
            columns ([str]): keys to access data in items and also the
              column names in the destination table.
            default_values (dict): dictionary of default values to use when
              inserting entried int the tblname table.
            cur: a db cursor; if not given, a new cursor will be created.
            item_cb (fn): optional function to apply to items's entry.
        """

        read_file, write_file = os.pipe()

        def writer():
            cursor = self.cursor(cur)
            with open(read_file, 'r') as f:
                cursor.copy_expert('COPY %s (%s) FROM STDIN CSV' % (
                    tblname, ', '.join(columns)), f)

        write_thread = threading.Thread(target=writer)
        write_thread.start()

        try:
            with open(write_file, 'w') as f:
                for d in items:
                    if item_cb is not None:
                        item_cb(d)
                    line = [escape(d.get(k, default_values.get(k)))
                            for k in columns]
                    f.write(','.join(line))
                    f.write('\n')

        finally:
            # No problem bubbling up exceptions, but we still need to make sure
            # we finish copying, even though we're probably going to cancel the
            # transaction.
            write_thread.join()

    def mktemp(self, tblname, cur=None):
        self.cursor(cur).execute('SELECT swh_mktemp(%s)', (tblname,))



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import multiprocessing
import os
import socket
import time
from urllib.request import urlopen

import aiohttp


class ServerTestFixtureBaseClass(metaclass=abc.ABCMeta):
    """Base class for http client/server testing implementations.

    Override this class to implement the following methods:
    - process_config: to do something needed for the server
      configuration (e.g propagate the configuration to other part)
    - define_worker_function: define the function that will actually
      run the server.

    To ensure test isolation, each test will run in a different server
    and a different folder.

    In order to correctly work, the subclass must call the parents
    class's setUp() and tearDown() methods.

    """
    def setUp(self):
        super().setUp()
        self.start_server()

    def tearDown(self):
        self.stop_server()
        super().tearDown()

    def url(self):
        return 'http://127.0.0.1:%d/' % self.port

    def process_config(self):
        """Process the server's configuration.  Do something useful for
        example, pass along the self.config dictionary inside the
        self.app.

        By default, do nothing.

        """
        pass

    @abc.abstractmethod
    def define_worker_function(self, app, port):
        """Define how the actual implementation server will run.

        """
        pass

    def start_server(self):
        """ Spawn the API server using multiprocessing.
        """
        self.process = None

        self.process_config()

        # Get an available port number
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('127.0.0.1', 0))
        self.port = sock.getsockname()[1]
        sock.close()

        worker_fn = self.define_worker_function()

        self.process = multiprocessing.Process(
            target=worker_fn, args=(self.app, self.port)
        )
        self.process.start()

        # Wait max 5 seconds for server to spawn
        i = 0
        while i < 500:
            try:
                urlopen(self.url())
            except Exception:
                i += 1
                time.sleep(0.01)
            else:
                return

    def stop_server(self):
        """ Terminate the API server's process.
        """
        if self.process:
            self.process.terminate()


class ServerTestFixture(ServerTestFixtureBaseClass):
    """Base class for http client/server testing (e.g flask).

    Mix this in a test class in order to have access to an http server
    running in background.

    Note that the subclass should define a dictionary in self.config
    that contains the server config.  And an application in self.app
    that corresponds to the type of server the tested client needs.

    To ensure test isolation, each test will run in a different server
    and a different folder.

    In order to correctly work, the subclass must call the parents
    class's setUp() and tearDown() methods.
    """
    def process_config(self):
        # WSGI app configuration
        for key, value in self.config.items():
            self.app.config[key] = value

    def define_worker_function(self):
        def worker(app, port):
            # Make Flask 1.0 stop printing its server banner
            os.environ['WERKZEUG_RUN_MAIN'] = 'true'
            return app.run(port=port, use_reloader=False)

        return worker


class ServerTestFixtureAsync(ServerTestFixtureBaseClass):
    """Base class for http client/server async testing (e.g aiohttp).

    Mix this in a test class in order to have access to an http server
    running in background.

    Note that the subclass should define an application in self.app
    that corresponds to the type of server the tested client needs.

    To ensure test isolation, each test will run in a different server
    and a different folder.

    In order to correctly work, the subclass must call the parents
    class's setUp() and tearDown() methods.

    """
    def define_worker_function(self):
        def worker(app, port):
            return aiohttp.web.run_app(app, port=int(port),
                                       print=lambda *_: None)

        return worker



#

import logging

import click
from click.testing import CliRunner

from swh.core.cli import swh as swhmain


help_msg = '''Usage: swh [OPTIONS] COMMAND [ARGS]...

  Command line interface for Software Heritage.

Options:
  -l, --log-level [NOTSET|DEBUG|INFO|WARNING|ERROR|CRITICAL]
                                  Log level (default to INFO)
  -h, --help                      Show this message and exit.
'''


def test_swh_help():
    runner = CliRunner()
    result = runner.invoke(swhmain, ['-h'])
    assert result.exit_code == 0
    assert result.output == help_msg

    result = runner.invoke(swhmain, ['--help'])
    assert result.exit_code == 0
    assert result.output == help_msg


def test_command():
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['test'])
    assert result.exit_code == 0
    assert result.output.strip() == 'Hello SWH!'


def test_loglevel_default(caplog):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 20
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['test'])
    assert result.exit_code == 0
    print(result.output)
    assert result.output.strip() == '''Hello SWH!'''


def test_loglevel_error(caplog):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 40
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['-l', 'ERROR', 'test'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''


def test_loglevel_debug(caplog):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 10
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['-l', 'DEBUG', 'test'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''


def test_aliased_command():
    @swhmain.command(name='canonical-test')
    @click.pass_context
    def swhtest(ctx):
        'A test command.'
        click.echo('Hello SWH!')
    swhmain.add_alias(swhtest, 'othername')

    runner = CliRunner()

    # check we have only 'canonical-test' listed in the usage help msg
    result = runner.invoke(swhmain, ['-h'])
    assert result.exit_code == 0
    assert 'canonical-test  A test command.' in result.output
    assert 'othername' not in result.output

    # check we can execute the cmd with 'canonical-test'
    result = runner.invoke(swhmain, ['canonical-test'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''

    # check we can also execute the cmd with the alias 'othername'
    result = runner.invoke(swhmain, ['othername'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Initially imported from https://github.com/DataDog/datadogpy/
# at revision 62b3a3e89988dc18d78c282fe3ff5d1813917436
#
# Copyright (c) 2015, Datadog <info@datadoghq.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Datadog nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from collections import deque
from contextlib import contextmanager
import os
import socket
import time
import unittest

import pytest

from swh.core.statsd import Statsd, TimedContextManagerDecorator


@contextmanager
def preserve_envvars(*envvars):
    """Context manager preserving the value of environment variables"""
    preserved = {}
    to_delete = object()

    for var in envvars:
        preserved[var] = os.environ.get(var, to_delete)

    yield

    for var in envvars:
        old = preserved[var]
        if old is not to_delete:
            os.environ[var] = old
        else:
            del os.environ[var]


class FakeSocket(object):
    """ A fake socket for testing. """

    def __init__(self):
        self.payloads = deque()

    def send(self, payload):
        assert type(payload) == bytes
        self.payloads.append(payload)

    def recv(self):
        try:
            return self.payloads.popleft().decode('utf-8')
        except IndexError:
            return None

    def close(self):
        pass

    def __repr__(self):
        return str(self.payloads)


class BrokenSocket(FakeSocket):
    def send(self, payload):
        raise socket.error("Socket error")


class SlowSocket(FakeSocket):
    def send(self, payload):
        raise socket.timeout("Socket timeout")


class TestStatsd(unittest.TestCase):

    def setUp(self):
        """
        Set up a default Statsd instance and mock the socket.
        """
        #
        self.statsd = Statsd()
        self.statsd.socket = FakeSocket()

    def recv(self):
        return self.statsd.socket.recv()

    def test_set(self):
        self.statsd.set('set', 123)
        assert self.recv() == 'set:123|s'

    def test_gauge(self):
        self.statsd.gauge('gauge', 123.4)
        assert self.recv() == 'gauge:123.4|g'

    def test_counter(self):
        self.statsd.increment('page.views')
        self.assertEqual('page.views:1|c', self.recv())

        self.statsd.increment('page.views', 11)
        self.assertEqual('page.views:11|c', self.recv())

        self.statsd.decrement('page.views')
        self.assertEqual('page.views:-1|c', self.recv())

        self.statsd.decrement('page.views', 12)
        self.assertEqual('page.views:-12|c', self.recv())

    def test_histogram(self):
        self.statsd.histogram('histo', 123.4)
        self.assertEqual('histo:123.4|h', self.recv())

    def test_tagged_gauge(self):
        self.statsd.gauge('gt', 123.4, tags={'country': 'china', 'age': 45})
        self.assertEqual('gt:123.4|g|#age:45,country:china', self.recv())

    def test_tagged_counter(self):
        self.statsd.increment('ct', tags={'country': 'españa'})
        self.assertEqual('ct:1|c|#country:españa', self.recv())

    def test_tagged_histogram(self):
        self.statsd.histogram('h', 1, tags={'test_tag': 'tag_value'})
        self.assertEqual('h:1|h|#test_tag:tag_value', self.recv())

    def test_sample_rate(self):
        self.statsd.increment('c', sample_rate=0)
        assert not self.recv()
        for i in range(10000):
            self.statsd.increment('sampled_counter', sample_rate=0.3)
        self.assert_almost_equal(3000, len(self.statsd.socket.payloads), 150)
        self.assertEqual('sampled_counter:1|c|@0.3', self.recv())

    def test_tags_and_samples(self):
        for i in range(100):
            self.statsd.gauge('gst', 23, tags={"sampled": True},
                              sample_rate=0.9)

        self.assert_almost_equal(90, len(self.statsd.socket.payloads), 10)
        self.assertEqual('gst:23|g|@0.9|#sampled:True', self.recv())

    def test_timing(self):
        self.statsd.timing('t', 123)
        self.assertEqual('t:123|ms', self.recv())

    def test_metric_namespace(self):
        """
        Namespace prefixes all metric names.
        """
        self.statsd.namespace = "foo"
        self.statsd.gauge('gauge', 123.4)
        self.assertEqual('foo.gauge:123.4|g', self.recv())

    # Test Client level contant tags
    def test_gauge_constant_tags(self):
        self.statsd.constant_tags = {
            'bar': 'baz',
        }
        self.statsd.gauge('gauge', 123.4)
        assert self.recv() == 'gauge:123.4|g|#bar:baz'

    def test_counter_constant_tag_with_metric_level_tags(self):
        self.statsd.constant_tags = {
            'bar': 'baz',
            'foo': True,
        }
        self.statsd.increment('page.views', tags={'extra': 'extra'})
        self.assertEqual(
            'page.views:1|c|#bar:baz,extra:extra,foo:True',
            self.recv(),
        )

    def test_gauge_constant_tags_with_metric_level_tags_twice(self):
        metric_level_tag = {'foo': 'bar'}
        self.statsd.constant_tags = {'bar': 'baz'}
        self.statsd.gauge('gauge', 123.4, tags=metric_level_tag)
        assert self.recv() == 'gauge:123.4|g|#bar:baz,foo:bar'

        # sending metrics multiple times with same metric-level tags
        # should not duplicate the tags being sent
        self.statsd.gauge('gauge', 123.4, tags=metric_level_tag)
        assert self.recv() == 'gauge:123.4|g|#bar:baz,foo:bar'

    def assert_almost_equal(self, a, b, delta):
        self.assertTrue(
            0 <= abs(a - b) <= delta,
            "%s - %s not within %s" % (a, b, delta)
        )

    def test_socket_error(self):
        self.statsd.socket = BrokenSocket()
        self.statsd.gauge('no error', 1)
        assert True, 'success'

    def test_socket_timeout(self):
        self.statsd.socket = SlowSocket()
        self.statsd.gauge('no error', 1)
        assert True, 'success'

    def test_timed(self):
        """
        Measure the distribution of a function's run time.
        """
        @self.statsd.timed('timed.test')
        def func(a, b, c=1, d=1):
            """docstring"""
            time.sleep(0.5)
            return (a, b, c, d)

        self.assertEqual('func', func.__name__)
        self.assertEqual('docstring', func.__doc__)

        result = func(1, 2, d=3)
        # Assert it handles args and kwargs correctly.
        self.assertEqual(result, (1, 2, 1, 3))

        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_timed_exception(self):
        """
        Exception bubble out of the decorator and is reported
        to statsd as a dedicated counter.
        """
        @self.statsd.timed('timed.test')
        def func(a, b, c=1, d=1):
            """docstring"""
            time.sleep(0.5)
            return (a / b, c, d)

        self.assertEqual('func', func.__name__)
        self.assertEqual('docstring', func.__doc__)

        with self.assertRaises(ZeroDivisionError):
            func(1, 0)

        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('c', type_)
        self.assertEqual('timed.test_error_count', name)
        self.assertEqual(int(value), 1)

    def test_timed_no_metric(self, ):
        """
        Test using a decorator without providing a metric.
        """

        @self.statsd.timed()
        def func(a, b, c=1, d=1):
            """docstring"""
            time.sleep(0.5)
            return (a, b, c, d)

        self.assertEqual('func', func.__name__)
        self.assertEqual('docstring', func.__doc__)

        result = func(1, 2, d=3)
        # Assert it handles args and kwargs correctly.
        self.assertEqual(result, (1, 2, 1, 3))

        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('swh.core.tests.test_statsd.func', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_timed_coroutine(self):
        """
        Measure the distribution of a coroutine function's run time.

        Warning: Python >= 3.5 only.
        """
        import asyncio

        @self.statsd.timed('timed.test')
        @asyncio.coroutine
        def print_foo():
            """docstring"""
            time.sleep(0.5)
            print("foo")

        loop = asyncio.get_event_loop()
        loop.run_until_complete(print_foo())
        loop.close()

        # Assert
        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_timed_context(self):
        """
        Measure the distribution of a context's run time.
        """
        # In milliseconds
        with self.statsd.timed('timed_context.test') as timer:
            self.assertIsInstance(timer, TimedContextManagerDecorator)
            time.sleep(0.5)

        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test', name)
        self.assert_almost_equal(500, float(value), 100)
        self.assert_almost_equal(500, timer.elapsed, 100)

    def test_timed_context_exception(self):
        """
        Exception bubbles out of the `timed` context manager and is
        reported to statsd as a dedicated counter.
        """
        class ContextException(Exception):
            pass

        def func(self):
            with self.statsd.timed('timed_context.test'):
                time.sleep(0.5)
                raise ContextException()

        # Ensure the exception was raised.
        self.assertRaises(ContextException, func, self)

        # Ensure the timing was recorded.
        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('c', type_)
        self.assertEqual('timed_context.test_error_count', name)
        self.assertEqual(int(value), 1)

    def test_timed_context_no_metric_name_exception(self):
        """Test that an exception occurs if using a context manager without a
        metric name.
        """

        def func(self):
            with self.statsd.timed():
                time.sleep(0.5)

        # Ensure the exception was raised.
        self.assertRaises(TypeError, func, self)

        # Ensure the timing was recorded.
        packet = self.recv()
        self.assertEqual(packet, None)

    def test_timed_start_stop_calls(self):
        timer = self.statsd.timed('timed_context.test')
        timer.start()
        time.sleep(0.5)
        timer.stop()

        packet = self.recv()
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_batched(self):
        self.statsd.open_buffer()
        self.statsd.gauge('page.views', 123)
        self.statsd.timing('timer', 123)
        self.statsd.close_buffer()

        self.assertEqual('page.views:123|g\ntimer:123|ms', self.recv())

    def test_context_manager(self):
        fake_socket = FakeSocket()
        with Statsd() as statsd:
            statsd.socket = fake_socket
            statsd.gauge('page.views', 123)
            statsd.timing('timer', 123)

        self.assertEqual('page.views:123|g\ntimer:123|ms', fake_socket.recv())

    def test_batched_buffer_autoflush(self):
        fake_socket = FakeSocket()
        with Statsd() as statsd:
            statsd.socket = fake_socket
            for i in range(51):
                statsd.increment('mycounter')
            self.assertEqual(
                '\n'.join(['mycounter:1|c' for i in range(50)]),
                fake_socket.recv(),
            )

        self.assertEqual('mycounter:1|c', fake_socket.recv())

    def test_module_level_instance(self):
        from swh.core.statsd import statsd
        self.assertTrue(isinstance(statsd, Statsd))

    def test_instantiating_does_not_connect(self):
        local_statsd = Statsd()
        self.assertEqual(None, local_statsd.socket)

    def test_accessing_socket_opens_socket(self):
        local_statsd = Statsd()
        try:
            self.assertIsNotNone(local_statsd.get_socket())
        finally:
            local_statsd.socket.close()

    def test_accessing_socket_multiple_times_returns_same_socket(self):
        local_statsd = Statsd()
        fresh_socket = FakeSocket()
        local_statsd.socket = fresh_socket
        self.assertEqual(fresh_socket, local_statsd.get_socket())
        self.assertNotEqual(FakeSocket(), local_statsd.get_socket())

    def test_tags_from_environment(self):
        with preserve_envvars('STATSD_TAGS'):
            os.environ['STATSD_TAGS'] = 'country:china,age:45'
            statsd = Statsd()

        statsd.socket = FakeSocket()
        statsd.gauge('gt', 123.4)
        self.assertEqual('gt:123.4|g|#age:45,country:china',
                         statsd.socket.recv())

    def test_tags_from_environment_and_constant(self):
        with preserve_envvars('STATSD_TAGS'):
            os.environ['STATSD_TAGS'] = 'country:china,age:45'
            statsd = Statsd(constant_tags={'country': 'canada'})
        statsd.socket = FakeSocket()
        statsd.gauge('gt', 123.4)
        self.assertEqual('gt:123.4|g|#age:45,country:canada',
                         statsd.socket.recv())

    def test_tags_from_environment_warning(self):
        with preserve_envvars('STATSD_TAGS'):
            os.environ['STATSD_TAGS'] = 'valid:tag,invalid_tag'
            with pytest.warns(UserWarning) as record:
                statsd = Statsd()

        assert len(record) == 1
        assert 'invalid_tag' in record[0].message.args[0]
        assert 'valid:tag' not in record[0].message.args[0]
        assert statsd.constant_tags == {'valid': 'tag'}

    def test_gauge_doesnt_send_none(self):
        self.statsd.gauge('metric', None)
        assert self.recv() is None

    def test_increment_doesnt_send_none(self):
        self.statsd.increment('metric', None)
        assert self.recv() is None

    def test_decrement_doesnt_send_none(self):
        self.statsd.decrement('metric', None)
        assert self.recv() is None

    def test_timing_doesnt_send_none(self):
        self.statsd.timing('metric', None)
        assert self.recv() is None

    def test_histogram_doesnt_send_none(self):
        self.statsd.histogram('metric', None)
        assert self.recv() is None

    def test_param_host(self):
        with preserve_envvars('STATSD_HOST', 'STATSD_PORT'):
            os.environ['STATSD_HOST'] = 'test-value'
            os.environ['STATSD_PORT'] = ''
            local_statsd = Statsd(host='actual-test-value')

        self.assertEqual(local_statsd.host, 'actual-test-value')
        self.assertEqual(local_statsd.port, 8125)

    def test_param_port(self):
        with preserve_envvars('STATSD_HOST', 'STATSD_PORT'):
            os.environ['STATSD_HOST'] = ''
            os.environ['STATSD_PORT'] = '12345'
            local_statsd = Statsd(port=4321)

        self.assertEqual(local_statsd.host, 'localhost')
        self.assertEqual(local_statsd.port, 4321)

    def test_envvar_host(self):
        with preserve_envvars('STATSD_HOST', 'STATSD_PORT'):
            os.environ['STATSD_HOST'] = 'test-value'
            os.environ['STATSD_PORT'] = ''
            local_statsd = Statsd()

        self.assertEqual(local_statsd.host, 'test-value')
        self.assertEqual(local_statsd.port, 8125)

    def test_envvar_port(self):
        with preserve_envvars('STATSD_HOST', 'STATSD_PORT'):
            os.environ['STATSD_HOST'] = ''
            os.environ['STATSD_PORT'] = '12345'
            local_statsd = Statsd()

        self.assertEqual(local_statsd.host, 'localhost')
        self.assertEqual(local_statsd.port, 12345)

    def test_namespace_added(self):
        local_statsd = Statsd(namespace='test-namespace')
        local_statsd.socket = FakeSocket()

        local_statsd.gauge('gauge', 123.4)
        assert local_statsd.socket.recv() == 'test-namespace.gauge:123.4|g'

    def test_contextmanager_empty(self):
        with self.statsd:
            assert True, 'success'

    def test_contextmanager_buffering(self):
        with self.statsd as s:
            s.gauge('gauge', 123.4)
            s.gauge('gauge_other', 456.78)
            self.assertIsNone(s.socket.recv())

        self.assertEqual(self.recv(), 'gauge:123.4|g\ngauge_other:456.78|g')

    def test_timed_elapsed(self):
        with self.statsd.timed('test_timer') as t:
            pass

        self.assertGreaterEqual(t.elapsed, 0)
        self.assertEqual(self.recv(), 'test_timer:%s|ms' % t.elapsed)



#!/usr/bin/env python3
# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import warnings
warnings.filterwarnings("ignore")  # noqa prevent psycopg from telling us sh*t

import click

from swh.core.cli import CONTEXT_SETTINGS

logger = logging.getLogger(__name__)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('module', nargs=-1, required=True)
@click.option('--db-name', '-d', help='Database name.',
              default='softwareheritage-dev', show_default=True)
def db_init(module, db_name=None):
    """Initialise a database for the Software Heritage <module>.  By
    default, attempts to create the database first.

    Example:

      swh db-init -d swh-test storage

    If you want to specify non-default postgresql connection parameters,
    please provide them using standard environment variables.
    See psql(1) man page (section ENVIRONMENTS) for details.

    Example:

      PGPORT=5434 swh db-init indexer

    """
    # put import statements here so we can keep startup time of the main swh
    # command as short as possible
    from os import path
    import glob
    from importlib import import_module
    from swh.core.utils import numfile_sortkey as sortkey
    from swh.core.tests.db_testing import (
        pg_createdb, pg_restore, DB_DUMP_TYPES,
        swh_db_version
    )

    logger.debug('db_init %s dn_name=%s', module, db_name)
    dump_files = []

    for modname in module:
        if not modname.startswith('swh.'):
            modname = 'swh.{}'.format(modname)
        try:
            m = import_module(modname)
        except ImportError:
            raise click.BadParameter(
                'Unable to load module {}'.format(modname))

        sqldir = path.join(path.dirname(m.__file__), 'sql')
        if not path.isdir(sqldir):
            raise click.BadParameter(
                'Module {} does not provide a db schema '
                '(no sql/ dir)'.format(modname))
        dump_files.extend(sorted(glob.glob(path.join(sqldir, '*.sql')),
                                 key=sortkey))

    # Create the db (or fail silently if already existing)
    pg_createdb(db_name, check=False)
    # Try to retrieve the db version if any
    db_version = swh_db_version(db_name)
    if not db_version:  # Initialize the db
        dump_files = [(x, DB_DUMP_TYPES[path.splitext(x)[1]])
                      for x in dump_files]
        for dump, dtype in dump_files:
            click.secho('Loading {}'.format(dump), fg='yellow')
            pg_restore(db_name, dump, dtype)

        db_version = swh_db_version(db_name)

    # TODO: Ideally migrate the version from db_version to the latest
    # db version

    click.secho('DONE database is {} version {}'.format(db_name, db_version),
                fg='green', bold=True)



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import stat
import tarfile
import zipfile

from os.path import abspath, realpath, join, dirname
from . import utils


def _canonical_abspath(path):
    """Resolve all paths to an absolute and real one.

    Args:
        path: to resolve

    Returns:
        canonical absolute path to path

    """
    return realpath(abspath(path))


def _badpath(path, basepath):
    """Determine if a path is outside basepath.

    Args:
        path: a relative or absolute path of a file or directory
        basepath: the basepath path must be in

    Returns:
        True if path is outside basepath, false otherwise.

    """
    return not _canonical_abspath(join(basepath, path)).startswith(basepath)


def _badlink(info, basepath):
    """Determine if the tarinfo member is outside basepath.

    Args:
        info: TarInfo member representing a symlink or hardlink of tar archive
        basepath: the basepath the info member must be in

    Returns:
        True if info is outside basepath, false otherwise.

    """
    tippath = _canonical_abspath(join(basepath, dirname(info.name)))
    return _badpath(info.linkname, basepath=tippath)


def is_tarball(filepath):
    """Given a filepath, determine if it represents an archive.

    Args:
        filepath: file to test for tarball property

    Returns:
        Bool, True if it's a tarball, False otherwise

    """
    return tarfile.is_tarfile(filepath) or zipfile.is_zipfile(filepath)


def _uncompress_zip(tarpath, dirpath):
    """Uncompress zip archive safely.

    As per zipfile is concerned
    (cf. note on https://docs.python.org/3.5/library/zipfile.html#zipfile.ZipFile.extract)  # noqa

    Args:
        tarpath: path to the archive
        dirpath: directory to uncompress the archive to

    """
    with zipfile.ZipFile(tarpath) as z:
        z.extractall(path=dirpath)


def _uncompress_tar(tarpath, dirpath):
    """Uncompress tarpath if the tarpath is safe.
    Safe means, no file will be uncompressed outside of dirpath.

    Args:
        tarpath: path to the archive
        dirpath: directory to uncompress the archive to

    Raises:
        ValueError when a member would be extracted outside dirpath.

    """
    def safemembers(tarpath, members, basepath):
        """Given a list of archive members, yield the members (directory,
        file, hard-link) that stays in bounds with basepath.  Note
        that symbolic link are authorized to point outside the
        basepath though.

        Args:
            tarpath: Name of the tarball
            members: Archive members for such tarball
            basepath: the basepath sandbox

        Yields:
            Safe TarInfo member

        Raises:
            ValueError when a member would be extracted outside basepath

        """
        errormsg = 'Archive {} blocked. Illegal path to %s %s'.format(tarpath)

        for finfo in members:
            if finfo.isdir() and _badpath(finfo.name, basepath):
                raise ValueError(errormsg % ('directory', finfo.name))
            elif finfo.isfile() and _badpath(finfo.name, basepath):
                raise ValueError(errormsg % ('file', finfo.name))
            elif finfo.islnk() and _badlink(finfo, basepath):
                raise ValueError(errormsg % ('hard-link', finfo.linkname))
            # Authorize symlinks to point outside basepath
            # elif finfo.issym() and _badlink(finfo, basepath):
            #     raise ValueError(errormsg % ('symlink', finfo.linkname))
            else:
                yield finfo

    with tarfile.open(tarpath) as t:
        members = t.getmembers()
        t.extractall(path=dirpath,
                     members=safemembers(tarpath, members, dirpath))


def uncompress(tarpath, dest):
    """Uncompress tarpath to dest folder if tarball is supported and safe.
       Safe means, no file will be uncompressed outside of dirpath.

       Note that this fixes permissions after successfully
       uncompressing the archive.

    Args:
        tarpath: path to tarball to uncompress
        dest: the destination folder where to uncompress the tarball

    Returns:
        The nature of the tarball, zip or tar.

    Raises:
        ValueError when:
        - an archive member would be extracted outside basepath
        - the archive is not supported

    """
    if tarfile.is_tarfile(tarpath):
        _uncompress_tar(tarpath, dest)
        nature = 'tar'
    elif zipfile.is_zipfile(tarpath):
        _uncompress_zip(tarpath, dest)
        nature = 'zip'
    else:
        raise ValueError('File %s is not a supported archive.' % tarpath)

    # Fix permissions
    for dirpath, _, fnames in os.walk(dest):
        os.chmod(dirpath, 0o755)
        for fname in fnames:
            fpath = os.path.join(dirpath, fname)
            if not os.path.islink(fpath):
                fpath_exec = os.stat(fpath).st_mode & stat.S_IXUSR
                if not fpath_exec:
                    os.chmod(fpath, 0o644)

    return nature


def _ls(rootdir):
    """Generator of filepath, filename from rootdir.

    """
    for dirpath, dirnames, fnames in os.walk(rootdir):
        for fname in (dirnames+fnames):
            fpath = os.path.join(dirpath, fname)
            fname = utils.commonname(rootdir, fpath)
            yield fpath, fname


def _compress_zip(tarpath, files):
    """Compress dirpath's content as tarpath.

    """
    with zipfile.ZipFile(tarpath, 'w') as z:
        for fpath, fname in files:
            z.write(fpath, arcname=fname)


def _compress_tar(tarpath, files):
    """Compress dirpath's content as tarpath.

    """
    with tarfile.open(tarpath, 'w:bz2') as t:
        for fpath, fname in files:
            t.add(fpath, arcname=fname, recursive=False)


def compress(tarpath, nature, dirpath_or_files):
    """Create a tarball tarpath with nature nature.
    The content of the tarball is either dirpath's content (if representing
    a directory path) or dirpath's iterable contents.

    Compress the directory dirpath's content to a tarball.
    The tarball being dumped at tarpath.
    The nature of the tarball is determined by the nature argument.

    """
    if isinstance(dirpath_or_files, str):
        files = _ls(dirpath_or_files)
    else:  # iterable of 'filepath, filename'
        files = dirpath_or_files

    if nature == 'zip':
        _compress_zip(tarpath, files)
    else:
        _compress_tar(tarpath, files)

    return tarpath



import json
import logging
import pickle
import sys
import traceback

import aiohttp.web
import multidict

from .serializers import msgpack_dumps, msgpack_loads, SWHJSONDecoder


def encode_data_server(data, **kwargs):
    return aiohttp.web.Response(
        body=msgpack_dumps(data),
        headers=multidict.MultiDict({'Content-Type': 'application/x-msgpack'}),
        **kwargs
    )


async def decode_request(request):
    content_type = request.headers.get('Content-Type')
    data = await request.read()
    if not data:
        return {}
    if content_type == 'application/x-msgpack':
        r = msgpack_loads(data)
    elif content_type == 'application/json':
        r = json.loads(data, cls=SWHJSONDecoder)
    else:
        raise ValueError('Wrong content type `%s` for API request'
                         % content_type)
    return r


async def error_middleware(app, handler):
    async def middleware_handler(request):
        try:
            return (await handler(request))
        except Exception as e:
            if isinstance(e, aiohttp.web.HTTPException):
                raise
            logging.exception(e)
            exception = traceback.format_exception(*sys.exc_info())
            res = {'exception': exception,
                   'exception_pickled': pickle.dumps(e)}
            return encode_data_server(res, status=500)
    return middleware_handler


class SWHRemoteAPI(aiohttp.web.Application):
    def __init__(self, *args, middlewares=(), **kwargs):
        middlewares = (error_middleware,) + middlewares
        super().__init__(*args, middlewares=middlewares, **kwargs)
